#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System;
using System.Collections.Generic;
using System.Threading;
using System.Linq;

namespace HeuristicLab.Optimization.Networks {
  [Item("AlgorithmServiceNode", "A node of an optimization network which provides a HeuristicLab algorithm as a service.")]
  [StorableClass]
  public class AlgorithmServiceNode : Node, IAlgorithmServiceNode {
    private object locker = new object();
    private Dictionary<IAlgorithm, AutoResetEvent> waitHandles = new Dictionary<IAlgorithm,AutoResetEvent>();

    new public PortCollection Ports {
      get { return base.Ports; }
    }

    [Storable]
    private IAlgorithm template;
    public IAlgorithm Template {
      get { return template; }
      set {
        if (value != template) {
          template = value;
          OnTemplateChanged();
        }
      }
    }

    [Storable]
    private RunCollection runs;
    public RunCollection Runs {
      get { return runs; }
    }

    [StorableConstructor]
    protected AlgorithmServiceNode(bool deserializing) : base(deserializing) { }
    protected AlgorithmServiceNode(AlgorithmServiceNode original, Cloner cloner)
      : base(original, cloner) {
      template = cloner.Clone(original.template);
      runs = cloner.Clone(original.runs);
    }
    public AlgorithmServiceNode()
      : base("AlgorithmServiceNode") {
      runs = new RunCollection();
    }
    public AlgorithmServiceNode(string name)
      : base(name) {
      runs = new RunCollection();
    }
    public AlgorithmServiceNode(string name, string description)
      : base(name, description) {
      runs = new RunCollection();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new AlgorithmServiceNode(this, cloner);
    }

    private void UpdateParameter(IInputPort port) {
      if (template != null) {
        IParameter param = null;
        if (template.Parameters.TryGetValue(port.Name, out param)) {
          IValueParameter p = param as IValueParameter;
          if ((p != null) && (port.Value != null)) {
            p.Value = (IItem)port.Value;
          }
        }
      }
    }

    protected virtual void Execute(ServiceParameterCollection parameters, CancellationToken token) {
      if (Template == null) throw new InvalidOperationException("Template is null");

      IAlgorithm algorithm;
      lock (locker) {
        algorithm = (IAlgorithm)Template.Clone();
        waitHandles.Add(algorithm, new AutoResetEvent(false));
      }

      // set parameters
      foreach (var p in parameters.Where(x => x.Type == ServiceParameterType.Input)) {
        IParameter param = null;
        if (algorithm.Parameters.TryGetValue(p.Name, out param)) {
          var v = param as IValueParameter;
          if (v != null) v.Value = p.Value;
        }
        if (algorithm.Problem.Parameters.TryGetValue(p.Name, out param)) {
          var v = param as IValueParameter;
          if (v != null) v.Value = p.Value;
        }
      }

      algorithm.StoreAlgorithmInEachRun = false;
      algorithm.Runs.Clear();
      algorithm.Prepare();
      algorithm.Stopped += Algorithm_Stopped;
      algorithm.Start();
      if (WaitHandle.WaitAny(new WaitHandle[] { waitHandles[algorithm], token.WaitHandle }) == 1) {
        // retrieve results
        foreach (var p in parameters.Where(x => x.Type == ServiceParameterType.Output)) {
          IResult result = null;
          if (algorithm.Results.TryGetValue(p.Name, out result)) {
            p.Value = result.Value;
          }
        }

        lock (locker) {
          waitHandles[algorithm].Dispose();
          waitHandles.Remove(algorithm);
          Runs.Add(algorithm.Runs.ToArray()[0]);
        }
      } else {  // cancellation
        algorithm.Stop();
        lock (locker) {
          waitHandles[algorithm].Dispose();
          waitHandles.Remove(algorithm);
        }
      }
    }

    private void Algorithm_Stopped(object sender, EventArgs e) {
      lock (locker) {
        waitHandles[(IAlgorithm)sender].Set();
      }
    }

    public event EventHandler TemplateChanged;
    protected virtual void OnTemplateChanged() {
      var handler = TemplateChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    #region Ports Events
    protected override void RegisterPortsEvents() {
      base.RegisterPortsEvents();
      foreach (var p in Ports)
        RegisterPortEvents(p);
    }
    protected override void Ports_ItemsAdded(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      base.Ports_ItemsAdded(sender, e);
      foreach (var p in e.Items)
        RegisterPortEvents(p);
    }
    protected override void Ports_ItemsRemoved(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      base.Ports_ItemsRemoved(sender, e);
      foreach (var p in e.Items)
        DeregisterPortEvents(p);
    }
    protected override void Ports_ItemsReplaced(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      base.Ports_ItemsReplaced(sender, e);
      foreach (var p in e.OldItems)
        DeregisterPortEvents(p);
      foreach (var p in e.Items)
        RegisterPortEvents(p);
    }
    protected override void Ports_CollectionReset(object sender, Collections.CollectionItemsChangedEventArgs<IPort> e) {
      base.Ports_CollectionReset(sender, e);
      foreach (var p in e.OldItems)
        DeregisterPortEvents(p);
      foreach (var p in e.Items)
        RegisterPortEvents(p);
    }
    #endregion

    #region Port Events
    private void RegisterPortEvents(IPort port) {
      IInputPort i = port as IInputPort;
      if (i != null) {
        i.NameChanged += InputPort_NameChanged;
        i.ValueChanged += InputPort_ValueChanged;
        UpdateParameter(i);
      }
      IServicePort s = port as IServicePort;
      if (s != null) {
        s.ProcessParameters += ServicePort_ProcessParameters;
      }
    }

    private void DeregisterPortEvents(IPort port) {
      IInputPort i = port as IInputPort;
      if (i != null) {
        i.NameChanged -= InputPort_NameChanged;
        i.ValueChanged -= InputPort_ValueChanged;
      }
      IServicePort s = port as IServicePort;
      if (s != null) {
        s.ProcessParameters -= ServicePort_ProcessParameters;
      }
    }
    private void InputPort_NameChanged(object sender, EventArgs e) {
      UpdateParameter((IInputPort)sender);
    }
    private void InputPort_ValueChanged(object sender, EventArgs e) {
      UpdateParameter((IInputPort)sender);
    }
    private void ServicePort_ProcessParameters(object sender, EventArgs<ServiceParameterCollection, CancellationToken> e) {
      Execute(e.Value, e.Value2);
    }
    #endregion
  }
}
